#include <PID_v1.h>
#include <AFMotor.h>

/*
 *  Program to drive Tilt platform with Arduino-Mega and Motor Shield L298:
 *  
 *  Author:  Florian Bauer - Jun 2023
 *  (there is free sourc and no copy right on this code, use it with fun without guarantee)
 *  
 *  2 Configurations possible:
 *  a) 2 Servos  
 *  b) 2 XM Motors with Quadrature Encoders
 *  
 *  Uncomment configuration in sketch below, compile and upload to Arduino-Mega
 *  #define ACTUATOR MOTORSERVO
 *  or
 *  #define ACTUATOR SERVO
 *  
 *  This sketch implements following Serial Communication Protocol:
 *  <x>,<y>   tilt platform to x- and y-angle, range -20..20
 *
 *  relevant if a DC-Motor is used as Servo
 *  kp<p>     set kp of Servo-PID to p
 *  ki<i>     set ki of Servo-PID to i
 *  kd<d>     set kd of Servo-PID to d
 *  l<l>      set limit of Servo-PID to l
 *  r         set actual position as zero position, reset encoder counters
 *  p         toggle print mode : if on print all parameters
 *  c         toggle control mode : if on Joystick-Control of platform
 *            Analog-Joystick connected to analog pins A2 and A3
 *            
 *  Serial speed is 250000 bd
 *            
 *     Motor Connections wiht Arduino:
 *     
 *     a) with Servos:    M2: Servo0, M1: Servo1
 *     b) with XM Motors: M1: Motor (encoderPin1,encoderPin2) , M2: Motor (encoderPin3,encoderPin4)  
 *                Connections from Motor Mx rear side:
 *                   +----------+
 *                   |          |
 *                   |          | 
 *                   |          |                    
 *                   | + 1234 - |  1=GND,2=VCC (5V), 
 *                   +----------+
 *     for Motor M1: 3=A (encoderPin2), 4=B (encoderPin1), +,-: Motor-Shield L298 M1
 *     for Motor M2: 3=A (encoderPin3), 4=B (encoderPin4)  +,-: Motor-Shield L298 M2 
 *     
 *     /!\ deviating wiring will lead to movements inconsistent with controller program
 *     
 *     Platform Movements from top view when submitting x,y command:   
 *     (Verify in Arduino-Console with x,y-Command!)
 *     
 *     0,10   M2 +----+       
 *               |    |
                 |   x|    right side goes down
 *               |    |
 *            M1 +----+         
 *      
 *     0,-10  M2 +----+       
 *               |    |
                 |   .|    right side goes up
 *               |    |
 *            M1 +----+  
 *      
 *      
 *     10,0   M2 +----+       
 *               | x  |
                 |    |    top goes down
 *               |    |
 *            M1 +----+  
 *      
 *      
 *    -10,0   M2 +----+       
 *               | ´. |
                 |    |    top goes up 
 *               |    |
 *            M1 +----+  
 *      
 * For different Motors, Servos the calibration parameters in the sketch may need adaptions    
 *   Parameters depend on Motor / Servo and steering geometry
 *   
 *   - Mechanically level platform horizontally
 *   
 *   - Measure different positions entered with x,y-Command and calculate fit parameters
 * 
 */
 
//#define SERIAL


#define SERIAL

#define MOTORSERVO 1
#define SERVO      0

//#define ACTUATOR MOTORSERVO
#define ACTUATOR SERVO

#if ACTUATOR == MOTORSERVO
 AF_DCMotor mot1(1,MOTOR12_2KHZ);
 AF_DCMotor mot2(2,MOTOR12_2KHZ);
#endif

#if ACTUATOR == SERVO
#include <Servo.h>              // Add library
Servo servo1,servo2;               // Define any servo name
#endif


String readString; //This while store the user input data
int User_Input = 0; // This while convert input string into integer
int abcd;
int User_Input2 = 0; // This while convert input string into integer
int printmode=0;
int controlmode=0;

int encoderPin1 = 18; //Encoder Output 'A' must connected with intreput pin of arduino.
int encoderPin2 = 19; //Encoder Otput 'B' must connected with intreput pin of arduino.
int encoderPin3 = 21; //Encoder Otput 'B' must connected with intreput pin of arduino.
int encoderPin4 = 20; //Encoder Otput 'B' must connected with intreput pin of arduino.


volatile int updn = 1;
volatile int updn2 = 1;

volatile int lastEncoded = 0; // Here updated value of encoder store.
volatile long encoderValue = 0; // Raw encoder value

volatile int lastEncoded2 = 0; // Here updated value of encoder store.
volatile long encoderValue2 = 0; // Raw encoder value

int PPR = 1600;  // Encoder Pulse per revolution.
int angle = 360; // Maximum degree of motion.
int REV = 0;          // Set point REQUIRED ENCODER VALUE
int REVX,REVY;
int REVA,REVB;
int loopcounter=0;

int lastMSB = 0;
int lastLSB = 0;
double kp = 250, ki = .0, kd = 0.3; //0.01;             // modify for optimal performance
 double inputOld=0, input = 0, output = 0, setpoint = 0;
  double input2Old=0, input2 = 0, output2 = 0, setpoint2 = 0;
int limit = 222;

   PID myPID  =   PID (&input, &output, &setpoint, kp, ki, kd, DIRECT);  
   PID myPID2 = PID(&input2, &output2, &setpoint2, kp, ki, kd, DIRECT);  

  
double x0=0.0, y0=0.0;
double xraw,yraw;

void setup() {
  pinMode(15,OUTPUT);
  digitalWrite(15,HIGH);

 // pinMode(MotEnable, OUTPUT);
  //pinMode(MotFwd, OUTPUT); 
  //pinMode(MotRev, OUTPUT); 
 //    pinMode(MotPWMC, OUTPUT);
 // pinMode(MotBrkC, OUTPUT);
 // pinMode(MotDirC, OUTPUT);

 //      pinMode(MotPWMA, OUTPUT);
 // pinMode(MotBrkA, OUTPUT);
 // pinMode(MotDirA, OUTPUT);
#ifdef SERIAL  
  Serial.begin(250000); //initialize serial comunication
  // TCCR0B = TCCR0B & 0b11111000 | 0b01;  // set 31KHz PWM to prevent motor noise
  // TCCR1B = TCCR1B & 0b11111000 | 0b01;  // set 31KHz PWM to prevent motor noise
  // TCCR2B = TCCR2B & 0b11111000 | 0b01;  // set 31KHz PWM to prevent motor noise
#endif

#if ACTUATOR == MOTORSERVO
pinMode(encoderPin1, INPUT_PULLUP); 
     digitalWrite(encoderPin1, HIGH); //turn pullup resistor on
attachInterrupt(digitalPinToInterrupt(encoderPin1), updateEncoder, CHANGE);


  pinMode(encoderPin2, INPUT_PULLUP);
  digitalWrite(encoderPin2, HIGH); //turn pullup resistor on
  attachInterrupt(digitalPinToInterrupt(encoderPin2), updateEncoder, CHANGE);




  pinMode(encoderPin3, INPUT_PULLUP);
  digitalWrite(encoderPin3, HIGH); //turn pullup resistor on
  attachInterrupt(digitalPinToInterrupt(encoderPin3), updateEncoder2, CHANGE);

    pinMode(encoderPin4, INPUT_PULLUP);
  digitalWrite(encoderPin4, HIGH); //turn pullup resistor on
  attachInterrupt(digitalPinToInterrupt(encoderPin4), updateEncoder2, CHANGE);
 
  //call updateEncoder() when any high/low changed seen
  //on interrupt 0 (pin 2), or interrupt 1 (pin 3) 


     
 //  TCCR0B = TCCR0B & 0b11111000 | 0b01;  // set 31KHz PWM to prevent motor noise
 //  TCCR1B = TCCR1B & 0b11111000 | 0b01;  // set 31KHz PWM to prevent motor noise
 //  TCCR2B = TCCR2B & 0b11111000 | 0b01;  // set 31KHz PWM to prevent motor noise


  myPID.SetMode(AUTOMATIC);   //set PID in Auto mode
  myPID.SetSampleTime(1);  // refresh rate of PID controller
  myPID.SetOutputLimits(-limit, limit); // this is the MAX PWM value to move motor, here change in value reflect change in speed of motor.

  myPID2.SetMode(AUTOMATIC);   //set PID in Auto mode
 myPID2.SetSampleTime(1);  // refresh rate of PID controller
  myPID2.SetOutputLimits(-limit, limit); // this is the MAX PWM value to move motor, here change in value reflect change in speed of motor.
#endif 

#if ACTUATOR == SERVO
  servo1.attach (10);          // Define the servo signal pins
  servo2.attach (9);          // Define the servo signal pins
#endif

}
 float x=0,y=0;
void loop() {
  char buf[64];
 float tmp;
  char *ptr;
#ifdef SERIAL

  
  while (Serial.available()) { //Check if the serial data is available.
    delay(3);                  // a small delay
    char c = Serial.read();  // storing input data
    readString += c;         // accumulate each of the characters in readString
  }
 
   if (readString.length() >0) { //Verify that the variable contains information  
      readString.toCharArray(buf, 100);
  
      if(strncmp(buf,"p",1)==0)
      { printmode=1-printmode; }
      else if(strncmp(buf,"c",1)==0) 
      { controlmode=1-controlmode; }
      else if(strncmp(buf,"r",1)==0) 
      { 
     
#if ACTUATOR == MOTORSERVO  
        encoderValue=0;
      encoderValue2=0;
      
      x=0.0;
      y=0.0;
#endif

#if ACTUATOR == SERVO 
      x0 =xraw;
      y0 =yraw;
     
#endif
      }
      else if(strncmp(buf,"l",1)==0) 
      { 
        limit = atoi(buf+1);
      myPID.SetOutputLimits(-limit, limit);
      myPID2.SetOutputLimits(-limit, limit);
      if(printmode) { Serial.print("Limit="); Serial.println(limit); }
      
      }
#if ACTUATOR == MOTORSERVO    
       else if(strncmp(buf,"kp",2)==0) 
      { 
        kp = atof(buf+2);
        myPID.SetTunings(kp,ki,kd);
        myPID2.SetTunings(kp,ki,kd);
        if(printmode) { Serial.print("kp="); Serial.println(kp); }
      }
      else if(strncmp(buf,"ki",2)==0) 
      { 
        ki = atof(buf+2);
        myPID.SetTunings(kp,ki,kd);
        myPID2.SetTunings(kp,ki,kd);
        if(printmode) { Serial.print("ki="); Serial.println(ki); }
      }
      else if(strncmp(buf,"kd",2)==0) 
      { 
        kd = atof(buf+2);
        myPID.SetTunings(kp,ki,kd);
        myPID2.SetTunings(kp,ki,kd);
        if(printmode) { Serial.print("kd="); Serial.println(kd); }
      }
#endif
      else 
      {
        ptr = strtok(buf, ",");
     
     // Serial.println(ptr);
      if (ptr != NULL) {
        x = -atof(ptr) + x0;
        xraw=x;
// Following Parameters depend on Motor / Servo and steering geometry
// Calibrate this by measuring different positions

#if ACTUATOR == MOTORSERVO
        x = x / 0.15;
#endif
#if ACTUATOR == SERVO
      x = x / 0.28 +80.0; 
#endif

        ptr = strtok(NULL, ",");
        if(ptr !=NULL)
        {   
          y = -atof(ptr) + y0;   // here input data is store in integer form
          yraw=y;
#if ACTUATOR == MOTORSERVO
          y = 10 *y;
#endif
#if ACTUATOR == SERVO
         // May need adustment to servo
           y = y / 0.43 + 90.0; 
          
#endif
        }
#if ACTUATOR == MOTORSERVO
        tmp=x;
        x=y;
        y=tmp;
#endif
    }
      }
    //Serial.print("SOLL:"); Serial.print(a); Serial.print(","); Serial.println(b);
    readString="";
}
#endif


 //Serial.print(User_Input);
 //Serial.print(",");
 //Serial.println(User_Input2);
 

if(controlmode)
{ User_Input = analogRead(2); // liest das Potentiometer aus (Wert zwischen 0 und 1023)
  User_Input2 = analogRead(3); // liest das Potentiometer aus (Wert zwischen 0 und 1023)
  REVX = map(User_Input, 0, 1023, -250, 250);
  REVY = map(User_Input2, 0, 1023, -250, 250);
}
else
{
  REVX = y;
 REVY = x;
}
  REVA=(REVX + REVY ) / 2;
  REVB=(REVY - REVX) / 2 ;

//REV = map (User_Input, 0, 360, 0, 800*3/8); // mapping degree into pulse

setpoint = REVA;     
setpoint2 = REVB;                    //PID while work to achive this value consider as SET value

#if ACTUATOR == MOTORSERVO
//cli();//PID while work to achive this value consider as SET value
input = encoderValue; // /3;           // data from encoder consider as a Process value
//sei();


myPID.Compute();                 // calculate new output
pwmOut(0,output);



//cli();
input2 =  encoderValue2 ; 
//sei();

myPID2.Compute();                 // calculate new output
pwmOut(1,output2);
#endif

#if ACTUATOR == SERVO
int servo_position1 = (int) (0.5 + ( 0.5* ( (float)x + (float)y)));                    //PID while work to achive this value consider as SET value
int servo_position2 = (int) (0.5 + (0.5 *( (float)x - (float)y +90)));                    //PID while work to achive this value consider as SET value
servo1.write(servo_position1);
servo2.write(servo_position2);
#endif

if(printmode && (input != inputOld || input2 != input2Old))
{
  Serial.print(REVA); Serial.print(" "); Serial.print(input); Serial.print(" ");
  Serial.print(REVB); Serial.print(" "); Serial.print(input2);Serial.print(" ");Serial.println(controlmode);
  
}
 inputOld=input;
 input2Old=input2;

if( abs(input2) < 2
)
digitalWrite(15,LOW);
else
digitalWrite(15,HIGH);

 


//pwmOut2(output2);  




 loopcounter++;

}

#if ACTUATOR == MOTORSERVO
void pwmOut(int i,int out) {     
                
  if (out > 0) {                         // if REV > encoderValue motor move in forward direction.    
     switch(i)
     {
      case 0: mot1.run(FORWARD); mot1.setSpeed(out); break;
       case 1: mot2.run(FORWARD); mot2.setSpeed(out); break;
     //   case 2: mot3.run(FORWARD); mot3.setSpeed(out); break;
     }
  }
  else {
     switch(i) {
     case 0 : mot1.run(BACKWARD); mot1.setSpeed(abs(out)); break;
      case 1: mot2.run(BACKWARD); mot2.setSpeed(abs(out)); break;
   //     case 2: mot3.run(BACKWARD); mot3.setSpeed(abs(out)); break;
     }

  }
}



void updateEncoder(){
  int MSB = digitalRead(encoderPin1); //MSB = most significant bit
  int LSB = digitalRead(encoderPin2); //LSB = least significant bit

  int encoded = (MSB << 1) |LSB; //converting the 2 pin value to single number
  int sum  = (lastEncoded << 2) | encoded; //adding it to the previous encoded value

  if(sum == 0b1101 || sum == 0b0100 || sum == 0b0010 || sum == 0b1011) encoderValue ++;
  if(sum == 0b1110 || sum == 0b0111 || sum == 0b0001 || sum == 0b1000) encoderValue --;

  lastEncoded = encoded; //store this value for next time


}


void updateEncoder2(){
  //cli();
  
  int MSB = digitalRead(encoderPin3); //MSB = most significant bit
  int LSB = digitalRead(encoderPin4); //LSB = least significant bit

  int encoded2 = (MSB << 1) |LSB; //converting the 2 pin value to single number
  int sum  = (lastEncoded2 << 2) | encoded2; //adding it to the previous encoded value

  if(sum == 0b1101 || sum == 0b0100 || sum == 0b0010 || sum == 0b1011) encoderValue2 ++;
  if(sum == 0b1110 || sum == 0b0111 || sum == 0b0001 || sum == 0b1000) encoderValue2 --;

  lastEncoded2 = encoded2; //store this value for next time
 // sei();


}

void updateEncoderSingle(){
  int MSB = digitalRead(encoderPin3); //MSB = most significant bit
 
    if (updn)
    encoderValue2 ++;
    else
    encoderValue2 --;
   
  lastEncoded2 = encoderValue2; //store this value for next time

}

#endif
